package net.knarcraft.minigames.util;

import net.knarcraft.knarlib.util.TabCompletionHelper;
import net.knarcraft.minigames.MiniGames;
import net.knarcraft.minigames.arena.Arena;
import net.knarcraft.minigames.arena.ArenaHandler;
import net.knarcraft.minigames.arena.EditablePropertyType;
import net.knarcraft.minigames.arena.dropper.DropperArenaEditableProperty;
import net.knarcraft.minigames.arena.parkour.ParkourArenaEditableProperty;
import org.bukkit.Bukkit;
import org.bukkit.Material;
import org.bukkit.Tag;
import org.bukkit.event.entity.EntityDamageEvent;
import org.bukkit.permissions.Permission;
import org.jetbrains.annotations.NotNull;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringJoiner;

/**
 * A helper-class for common tab-completions
 */
public final class TabCompleteHelper {

    private static Map<EditablePropertyType, List<String>> tabCompleteSuggestions;
    private static List<String> plugins;
    private static Map<String, List<String>> permissions;

    private TabCompleteHelper() {

    }

    /**
     * Gets the names of all current arenas
     *
     * @return <p>All arena names</p>
     */
    @NotNull
    public static List<String> getDropperArenas() {
        return getArenas(MiniGames.getInstance().getDropperArenaHandler());
    }

    /**
     * Gets the names of all current arenas
     *
     * @return <p>All arena names</p>
     */
    @NotNull
    public static List<String> getParkourArenas() {
        return getArenas(MiniGames.getInstance().getParkourArenaHandler());
    }

    /**
     * Gets the names of all current arenas
     *
     * @return <p>All arena names</p>
     */
    @NotNull
    private static List<String> getArenas(@NotNull ArenaHandler<?, ?> arenaHandler) {
        List<String> arenaNames = new ArrayList<>();
        for (Arena arena : arenaHandler.getArenas().values()) {
            arenaNames.add(arena.getArenaName());
        }
        return arenaNames;
    }

    /**
     * Gets the argument strings of all arena properties
     *
     * @return <p>All arena properties</p>
     */
    @NotNull
    public static List<String> getDropperArenaProperties() {
        List<String> arenaProperties = new ArrayList<>();
        for (DropperArenaEditableProperty property : DropperArenaEditableProperty.values()) {
            arenaProperties.add(property.getArgumentString());
        }
        return arenaProperties;
    }

    /**
     * Gets the argument strings of all arena properties
     *
     * @return <p>All arena properties</p>
     */
    @NotNull
    public static List<String> getParkourArenaProperties() {
        List<String> arenaProperties = new ArrayList<>();
        for (ParkourArenaEditableProperty property : ParkourArenaEditableProperty.values()) {
            arenaProperties.add(property.getArgumentString());
        }
        return arenaProperties;
    }

    /**
     * Gets tab-complete suggestions for the given property type
     *
     * @param propertyType <p>The property type to get suggestions for</p>
     * @return <p>The suggestions produced</p>
     */
    @NotNull
    public static List<String> getTabCompleteSuggestions(@NotNull EditablePropertyType propertyType) {
        if (tabCompleteSuggestions == null) {
            tabCompleteSuggestions = new HashMap<>();
            tabCompleteSuggestions.put(EditablePropertyType.LOCATION, getLocationSuggestions());
            tabCompleteSuggestions.put(EditablePropertyType.ARENA_NAME, getNameSuggestions());
            tabCompleteSuggestions.put(EditablePropertyType.HORIZONTAL_VELOCITY, getHorizontalVelocitySuggestions());
            tabCompleteSuggestions.put(EditablePropertyType.VERTICAL_VELOCITY, getVerticalVelocitySuggestions());
            tabCompleteSuggestions.put(EditablePropertyType.BLOCK_TYPE, getBlockTypeSuggestions());
            tabCompleteSuggestions.put(EditablePropertyType.CHECKPOINT_CLEAR, getCheckpointClearSuggestions());
            tabCompleteSuggestions.put(EditablePropertyType.MATERIAL_LIST, getMaterialListSuggestions());
            tabCompleteSuggestions.put(EditablePropertyType.DOUBLE, getDoubleSuggestions());
            tabCompleteSuggestions.put(EditablePropertyType.INTEGER, getIntegerSuggestions());
            tabCompleteSuggestions.put(EditablePropertyType.DAMAGE_CAUSE_LIST, getDamageCauseSuggestions());
        }

        return tabCompleteSuggestions.get(propertyType);
    }

    /**
     * Gets tab-complete suggestions for list properties
     *
     * @param propertyType <p>Gets the type of property to show tab-completions for</p>
     * @param input        <p>The input string to use for filtering</p>
     * @return <p>The tab-completions to suggest</p>
     */
    @NotNull
    public static List<String> getListCompleteSuggestions(@NotNull EditablePropertyType propertyType,
                                                          @NotNull String input) {
        List<String> allValues;
        if (propertyType == EditablePropertyType.DAMAGE_CAUSE_LIST) {
            allValues = getTabCompleteSuggestions((EditablePropertyType.DAMAGE_CAUSE_LIST));
        } else if (propertyType == EditablePropertyType.MATERIAL_LIST) {
            allValues = getTabCompleteSuggestions((EditablePropertyType.MATERIAL_LIST));
        } else {
            throw new IllegalArgumentException("Invalid property type given!");
        }
        return TabCompletionHelper.getStringList(allValues, input, TabCompletionHelper::filterMatchingContains);
    }

    /**
     * Gets tab-complete suggestions for a list of damage causes
     *
     * @return <p>Damage cause suggestions</p>
     */
    @NotNull
    private static List<String> getDamageCauseSuggestions() {
        List<String> suggestions = new ArrayList<>();
        for (EntityDamageEvent.DamageCause damageCause : EntityDamageEvent.DamageCause.values()) {
            suggestions.add(damageCause.name());
        }
        return suggestions;
    }

    /**
     * Gets tab-complete suggestions for an integer value
     *
     * @return <p>Integer suggestions</p>
     */
    @NotNull
    private static List<String> getIntegerSuggestions() {
        List<String> suggestions = new ArrayList<>();
        suggestions.add("-1");
        suggestions.add("1");
        suggestions.add("2");
        suggestions.add("3");
        suggestions.add("4");
        suggestions.add("5");
        suggestions.add("6");
        suggestions.add("7");
        return suggestions;
    }

    /**
     * Gets suggestions for double values
     *
     * @return <p>A list of suggestions</p>
     */
    @NotNull
    private static List<String> getDoubleSuggestions() {
        List<String> suggestions = new ArrayList<>();
        suggestions.add("0");
        suggestions.add("0.01");
        suggestions.add("0.1");
        suggestions.add("0.2");
        suggestions.add("0.3");
        suggestions.add("0.4");
        suggestions.add("0.5");
        suggestions.add("1");
        return suggestions;
    }

    /**
     * Gets suggestions for a list of materials
     *
     * @return <p>A list of suggestions</p>
     */
    @NotNull
    private static List<String> getMaterialListSuggestions() {
        List<String> suggestions = new ArrayList<>();
        for (Material material : Material.values()) {
            if (material.isBlock()) {
                suggestions.add(material.getKey().getKey());
            }
        }
        for (Tag<Material> tag : getTags()) {
            suggestions.add("+" + tag.getKey().getKey());
        }
        return suggestions;
    }

    /**
     * Gets suggestions for checkpointClear
     *
     * @return <p>A list of suggestions</p>
     */
    @NotNull
    private static List<String> getCheckpointClearSuggestions() {
        List<String> suggestions = new ArrayList<>();
        suggestions.add("true");
        return suggestions;
    }

    /**
     * Gets suggestions for a block material
     *
     * @return <p>A list of suggestions</p>
     */
    @NotNull
    private static List<String> getBlockTypeSuggestions() {
        List<String> materials = new ArrayList<>();
        for (Material material : Material.values()) {
            if (material.isBlock()) {
                materials.add(material.name());
            }
        }
        return materials;
    }

    /**
     * Gets suggestions for a vertical velocity
     *
     * @return <p>A list of suggestions</p>
     */
    @NotNull
    private static List<String> getVerticalVelocitySuggestions() {
        List<String> velocities = new ArrayList<>();
        velocities.add("0.01");
        velocities.add("0.5");
        velocities.add("1");
        velocities.add("3");
        velocities.add("10");
        velocities.add("30");
        velocities.add("75");
        return velocities;
    }

    /**
     * Gets suggestions for a horizontal velocity
     *
     * @return <p>A list of suggestions</p>
     */
    @NotNull
    private static List<String> getHorizontalVelocitySuggestions() {
        List<String> velocities = new ArrayList<>();
        velocities.add("0.01");
        velocities.add("0.1");
        velocities.add("0.5");
        velocities.add("1");
        return velocities;
    }

    /**
     * Gets suggestions for an arena name
     *
     * @return <p>A list of suggestions</p>
     */
    @NotNull
    private static List<String> getNameSuggestions() {
        List<String> locations = new ArrayList<>();
        locations.add("DropperArena1");
        locations.add("DropperArena2");
        locations.add("ParkourArena1");
        locations.add("ParkourArena2");
        return locations;
    }

    /**
     * Gets suggestions for a location
     *
     * @return <p>A list of suggestions</p>
     */
    @NotNull
    private static List<String> getLocationSuggestions() {
        List<String> locations = new ArrayList<>();
        locations.add("here");
        locations.add("x,y,z");
        locations.add("null");
        return locations;
    }

    /**
     * Gets the tab complete value for the permission typed
     *
     * @param typedNode <p>The full permission node typed by the player</p>
     * @return <p>All known valid auto-complete options</p>
     */
    @NotNull
    public static List<String> tabCompletePermission(@NotNull String typedNode) {
        if (plugins == null) {
            loadAvailablePermissions();
        }
        List<String> output;
        if (typedNode.contains(".")) {
            List<String> matchingPermissions = permissions.get(typedNode.substring(0, typedNode.lastIndexOf(".")));
            if (matchingPermissions == null) {
                output = new ArrayList<>();
            } else {
                //Filter by the typed text
                output = TabCompletionHelper.filterMatchingStartsWith(matchingPermissions, typedNode);
            }
        } else {
            output = plugins;
        }

        //Add previous permissions in the comma-separated lists as a prefix
        return output;
    }

    /**
     * Loads all permissions available from bukkit plugins
     */
    private static void loadAvailablePermissions() {
        plugins = new ArrayList<>();
        permissions = new HashMap<>();

        for (Permission permission : Bukkit.getPluginManager().getPermissions()) {
            loadPermission(permission.getName());
        }
    }

    /**
     * Loads a given permission into the proper lists and maps
     *
     * @param permissionName <p>The permission to load</p>
     */
    private static void loadPermission(@NotNull String permissionName) {
        String[] permissionParts = permissionName.split("\\.");
        if (permissionParts.length == 1 && !plugins.contains(permissionParts[0])) {
            plugins.add(permissionParts[0]);
        } else if (permissionParts.length > 1) {
            StringJoiner pathJoiner = new StringJoiner(".");
            for (int j = 0; j < permissionParts.length - 1; j++) {
                pathJoiner.add(permissionParts[j]);
            }
            String path = pathJoiner.toString();
            List<String> permissionList = permissions.computeIfAbsent(path, k -> new ArrayList<>());
            permissionList.add(permissionName);

            loadPermission(path);
        }
    }

    /**
     * Gets all useful tags for physical blocks
     *
     * @return <p>All block tags</p>
     */
    @NotNull
    private static Set<Tag<Material>> getTags() {
        // TODO: Find a way to get the tags programmatically
        return Set.of(Tag.ACACIA_LOGS, Tag.ALL_HANGING_SIGNS, Tag.ALL_SIGNS, Tag.ANVIL, Tag.WALLS,
                Tag.BAMBOO_BLOCKS, Tag.BANNERS, Tag.BEDS, Tag.BEEHIVES, Tag.BIRCH_LOGS,
                Tag.BUTTONS, Tag.CAMPFIRES, Tag.CANDLE_CAKES, Tag.CANDLES, Tag.CAULDRONS, Tag.CAVE_VINES,
                Tag.CEILING_HANGING_SIGNS, Tag.CHERRY_LOGS, Tag.CLIMBABLE, Tag.COAL_ORES, Tag.CONCRETE_POWDER,
                Tag.COPPER_ORES, Tag.CORAL_BLOCKS, Tag.CORAL_PLANTS, Tag.CORALS, Tag.CRIMSON_STEMS, Tag.CROPS,
                Tag.DARK_OAK_LOGS, Tag.DIAMOND_ORES, Tag.DIRT, Tag.DOORS, Tag.EMERALD_ORES, Tag.FENCE_GATES, Tag.FENCES,
                Tag.FIRE, Tag.FLOWER_POTS, Tag.FLOWERS, Tag.GOLD_ORES, Tag.ICE, Tag.IRON_ORES, Tag.JUNGLE_LOGS,
                Tag.LAPIS_ORES, Tag.LEAVES, Tag.LOGS, Tag.MANGROVE_LOGS, Tag.NYLIUM, Tag.OAK_LOGS, Tag.PLANKS,
                Tag.PORTALS, Tag.PRESSURE_PLATES, Tag.RAILS, Tag.REDSTONE_ORES, Tag.SAND, Tag.SAPLINGS, Tag.WART_BLOCKS,
                Tag.SHULKER_BOXES, Tag.SIGNS, Tag.SLABS, Tag.SMALL_FLOWERS, Tag.SNOW, Tag.SPRUCE_LOGS, Tag.STAIRS,
                Tag.STANDING_SIGNS, Tag.STONE_BRICKS, Tag.STONE_BUTTONS, Tag.TALL_FLOWERS, Tag.TERRACOTTA, Tag.WOOL,
                Tag.TRAPDOORS, Tag.WALL_CORALS, Tag.WALL_HANGING_SIGNS, Tag.WALL_SIGNS, Tag.WARPED_STEMS,
                Tag.WOODEN_BUTTONS, Tag.WOODEN_DOORS, Tag.WOODEN_FENCES, Tag.WOODEN_PRESSURE_PLATES, Tag.WOODEN_STAIRS,
                Tag.WOODEN_TRAPDOORS);
    }

}
